'''
	
	make_mmc_music_image.py

	OVERVIEW of Project:

	MMC cards were the precursor to SD cards.

	This file makes an MMC card image, that will have 16 song files for the Somerset Entertainment store kiosk machine.

	The file that this code outputs is based on a hex dump from the MMC card that came in the machine I purchased.

	I am not affliated with Somerset Ent. (Now defunct), and I reversed engineer as much of the file as I could, though not everything.

	This code builds a binary file that can be DD'd (*nix command DD) to an MMC card, so that it will play in the machine correctly.

	The final binary is made out of thee important parts: the header, the song data, and the footer.

	The HEADER:
		The header includes a date-stamp, some binary I am unsure of it's purpose, and a table of songs.
		This table of songs is very important - it stores the locations of the songs, their length and their name.
		Without this talbe, the machine will not be able to find the MP3s on the MMC card. Further, without the lengths
		the machine will not know how long to play the song.

		In addition to the songs we want to play, the machine requires 7 additional tracks (more about this in the footer).
		This 7 additional tracks must also be updated in the table.

		The header is always exactly 3072 bytes long, and the song index table always starts at byte 1024.

	The SONG DATA:
		The song data is very very simple. It is 16 mp3's back-to-back.

		That's it - just open the mp3s, and concatenate their binary back to back.

		Of course, while this song data is built we need to keep track of the song offsets, their names, and their lengths
		so the song index table can be updated in the header properly.

	The FOOTER:
		The footer is simply the factory messages for configuring the machine. This footer will be appended onto the final
		output UNCHANGED. It's simply 7 mp3s back-to-back, the first 6 being messages and the last being silence.

		Because our SONG DATA block will very in length, the offsets for the messages in the footer will need to be updated
		in the HEADERs table. But the footers data doesn't need to be changed at all.

	The goal of this code is simple:

		- Take a folder full of 16 mp3s, and sort them alphabetically (or numerically)
		- Open each mp3, and append them back to back, whilst making not about their position, length, and name.
		- Open the 3072 byte header block ripped from the original MMC card
		- Update the song index table in the header block with new offsets, song lengths, song names, and a date time-stamp.
		- Leave all the unknown binary in the header block unchanged.
		- Open up the footer file.
		- Build the final output, by appending the (modified) header data, the song data, and the footer data.
		- Save the final output to a binary file that can be DD'd to an MMC card.

	USAGE

	To use this script, run it with python on the command line.
	No non-standard libraries are required.

	Parameters:

	python make_mmc_music_image.py [folder_name] [output_filename]

		folder_name: 	A path, or a name of a folder to look for 16 MP3s in.
						This parameter is OPTIONAL, and will default to ./songs/

		output_filename:	A path / filename to save the MMC card image when done.
							This parameter is OPTIONAL, and will default to ./mmc_image

	TIPS:
		Some modern MP3s will have addionary binary inside of them, such as Album Art images, extra meta-data, and etc.
		The machine will error if unaccounted for data is found present (most likely hanging the machine.)
		It's therefore best to sanitze the MP3s before running this script.
		You can do this by converting them FROM MP3 TO MP3. Though pointless, it depending on your convertor it will most likely output uniform bare-bones MP3s.

		Also, many downloaded MP3s from the Inernet are mastered to different volume levels.
		You may want to normalize their volume before hand, so the machine plays them at a consistant level.
		This is especially important when the machine is on demo-mode.

	ETC:

		You may notice I have eccentricities in my programming style, such as:
		- I comment a metric shit-ton
		- I like large, screen width dividers for important areas of code
		- I close functions and loops with comments, because I don't like Pythons lack of { }

		Deal with it 8)

'''

# IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS
# IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS +++ IMPORTS

# We need OS so we can read files from the folder of songs
import os

# We need sys to read command line arguments for this script.
import sys

# We need math for math.floor, while calculating fractions of file offsets.
import math

# We need time so we can update the date in the final file.
import time



# SUPPORT FUNCTIONS +++ SUPPORT FUNCTIONS +++ SUPPORT FUNCTIONS +++ SUPPORT FUNCTIONS +++ SUPPORT FUNCTIONS +++ SUPPORT FUNCTIONS +++ SUPPORT FUNCTIONS +++ SUPPORT FUNCTIONS +++ SUPPORT FUNCTIONS
# SUPPORT FUNCTIONS +++ SUPPORT FUNCTIONS +++ SUPPORT FUNCTIONS +++ SUPPORT FUNCTIONS +++ SUPPORT FUNCTIONS +++ SUPPORT FUNCTIONS +++ SUPPORT FUNCTIONS +++ SUPPORT FUNCTIONS +++ SUPPORT FUNCTIONS

# convert an integer into a list of four-bytes that represent that integer
def getIntBytes(num):

	# Breif overview of this function:
	# We will convert the integer to a hex STRING, with leading zeros.
	# Then we well loop over the string 2 characters at a time, and treat those as a single hex byte.
	# We will then convert the single hex byte to a char.
	# The final output will be an array of four chars, which each represent the four bytes of the Integer.

	# Convert the number to a hex string.
	numInHex = "%08x" % num

	# Convert hex string into pairs of hex bytes as strings (using a list comprehension).
	bytesAsHexStr = [numInHex[i:(i+2)] for i in range(0, 8, 2) ]

	# Convert the hex bytes into Integer bytes.
	bytesAsChars = [chr(int( c, 16 )) for c in bytesAsHexStr]

	# Return the array of byte characters derived from an integer.
	return bytesAsChars
# }


# Write a list to specific position in a larger list, overwriting the data that was previously there.
# (this function mutates the larger list)
def writeListAtPos(outList, writeList, pos):

	# Loop over the list data to write.
	for i, v in enumerate(writeList):

		# Replace the existing data with the new value from the list we're writing.
		outList[pos+i] = v
    
    # Alternate syntax below (commented out), that uses more memory, but mildly faster.
    # Screw it, it's not like this thing runs slow.
    #[outList[pos+i] = v for (i,v) in enumerate(writeList)]

# }



# BEGINNING OF PROGRAM +++ BEGINNING OF PROGRAM +++ BEGINNING OF PROGRAM +++ BEGINNING OF PROGRAM +++ BEGINNING OF PROGRAM +++ BEGINNING OF PROGRAM +++ BEGINNING OF PROGRAM +++ BEGINNING OF PROGRAM
# BEGINNING OF PROGRAM +++ BEGINNING OF PROGRAM +++ BEGINNING OF PROGRAM +++ BEGINNING OF PROGRAM +++ BEGINNING OF PROGRAM +++ BEGINNING OF PROGRAM +++ BEGINNING OF PROGRAM +++ BEGINNING OF PROGRAM

'''
	OVERVIEW:
	we will get a folder (or use the default) full of 16 mp3s, sorted alphabetically
	we will get a filename to write the final binary MMC card image to
	we will loop over all the MP3s, read their binary data in, and concatenate them into one big binary blob
	whilst we are importing the songs, we will take note of their length in bytes, file name (truncated to 16 characters), and their offset in the blog
	we will import the header stripped from the original MMC card that came with the machine
	we will edit the original header to have new offsets, song lengths, and file names
	we will also edit the original header to have a new date stamp
	we will then attach the modified header (with song index table) to the blob of song's we built earlier
	we will write the entire concatenated data (header + songblob) to the output file
	we now have a working image

	TODO: Extra Credit

	Although not necessary, the original MMC card had "bonus" tracks for the programming menu.
	These bonus tracks will always have fixed relative offsets
	So we should open the bonus tracks (stripped from the end of the orginal MMC image)
	then append them to the end of the song blob
	and finally adjust the table so it's bonus track messages have offsets that match the new file

	NOTE: the extra credit has not yet been done, so the programming menu on the machine will error
'''

# Let's begin!

# Get the folder the songs are stored in if the user passed a directory,
# or default to 'songs/'
try:
	songFolder = sys.argv[1]
except:
	songFolder = 'songs'

# Let the user know which folder we're attempting to list the songs from.
print "\nReading Songs in " + songFolder + "/\n"

# Get the output file name the user passed in...
# or default to 'mmc_image'
try:
	outputFileName = sys.argv[2]
except:
	outputFileName = 'mmc_image'

# Get a list of the files in the folder we're going to be building into the MMC Image.
songFileList = os.listdir(songFolder + "/")

# We will be appending all the songs in one heap, so we can write everything all at once in the end.
# songBlob will be the block of binary MP3s stored back-to-back.
songBlob = ""

# We will store the global / local offsets, song lengths, and names / truncated names in an array of dictionaries.
# Basically all the important information about the songs will be stored in this array of dictionarys we're about to build.
songDetails = []

# Keep a running total of offsets (starting at 0, then adding the length of each file as we process them).
currentOffset = 0

# Loop through the files and open each one, determine it's properties, and append it to the larger blob of song data.
for songFileName in songFileList:

	# Let the user know we're reading in a song...
	print "Reading in song \"" + songFileName + "\""

	# Build the path to the file to open.
	songPathName = songFolder + "/" + songFileName;

	# Open the file and read it in.
	f = open(songPathName, 'rb')
	fData = f.read()
	f.close()

	# Add the data to the blob of songs.
	songBlob += fData

	# Get song info.

	# Truncate the file name and add a leading space.
	# Not sure if the leading space is necessary, but the original MMC card dump did this.
	songName16 = " " + songFileName.replace(".mp3","")[0:15]

	# Get song length.
	songLength = len(fData)

	# Save the current local offset.
	# (local meaning, local to the blob, where the first file is at position 0)
	songLocOffset = currentOffset

	# Save the global offset as well.
	# (the header will always be 3072 bytes long, so we can just add 3072 to the local offset)
	songGlobOffset = songLocOffset + 3072

	# Add the length of the song to the current offset, so the next song has a correct offset.
	currentOffset += songLength

	# Save the song details as a dictionary in the details array.
	songDetails += [{
						"name": songFileName,
						"name16": songName16,
						"length": songLength,
						"locOffset": songLocOffset,
						"globOffset": songGlobOffset
					}]

#} NEXT songFileName

# Now that we've added all the songs into one big block of data, let's measure the total size of this block of data.
songBlobSize = len(songBlob)

# Now we need to calculate the offset of the footer (we will need this in just a bit).
# The header is always exactly 3072 bytes, so the footer will begin at the end of the song blob.
# So the footer will begin at 3072 + songBlobSize
footerOffset = 3072 + songBlobSize

'''
	To recap:

	We've opened all of the files and concatenated them into one big chunk of data.
	Whilst we were opening the files, we populated an array of dictionaries containing their details.
	The details we kept about the songs are:
	- their full file name
	- their truncated name (16 bytes worth), with stripped ".mp3" extension
	- the total length of the song in bytes
	- their RELATIVE offset to the blob (first song = 0)
	- their GLOBAL offset (since we can assume the header will always be exactly 3072 bytes long)

	ANYHOO the next step is to build the song index table in the header.

	Instead of building it from scratch, we'll load a sample in from a file.

	The sample we're loading in was the original 3072 bytes ripped from the MMC card that came with the machine.
	Then we'll replace just the bytes we need to (the song offsets, lengths and names, etc).
'''

# Open the header file and read it's data.
headerFile = open('mmc_data/00-Header', 'rb')
headerData = headerFile.read()
headerFile.close()

# We need to convert the header data to a list so we can edit it on a byte-level basis later.
headerData = list(headerData)

'''
	Now we need to loop over the song data and replace it in the header data in the correct spots.
	
	The pattern for the song data is simple:
	- the data for the song index table starts at byte 1024 in the file (nice round number!)
	- each song is 32 bytes long
	- therefore, the song index table contains several 32 byte song entries back-to-back

	- the first four bytes are the song offset in the global file (so the first song would START at 3072 because the header is always 3072 long)
	- the next four bytes of the song data are length of the song in bytes, easy!
	- the next 8 bytes I'm not sure what they are, but here's some info anyway:
		the beginning of the next 8 starts with an ascii characater, either M or S. The next 2 bytes are some kind of numbers.
	  	each song in the file has three "sample points," for 0/3, 1/3 and 2/3 though the song.
	  	each sample point is actually three separate enteries in the header table
	  	the first entry for a song begins with M, where as the second two sample points (1/3 and 2/3 through the song) start with S
	  	the rest of the 8 bytes are some ascii numbers. Maybe the product SKU numbers? ISBNs? No idea.

	ANYWAY... since I'm not sure what those bytes are, we will leave the next 8 unchanged.

	After those 8, the next 16 bytes are the name of the song.

	Over all, each song has three entries in the header table, one for each third of the song (0/3, 1/2, and 2/3).
	There is no 3/3 because that would be the end of the song. The song can only start at 0%, 33% through, or 66% through.
	each entry looks like this:

	| 4 bytes | 4 bytes | 8 bytes         | 16 bytes                    |
	| Offset  | Length  | Unknown         | The name of the song        |


	So our next step is to start from 1024, and replace the first 4, second 4, and last 16 bytes of each 32 byte song block...
'''

# Let the user know we're attempting to populate the table.
print "\nBuilding the song index table...\n"

# Keep track of the position in the file for our current song, starting at the first entry which is always at 1024.
filePos = 1024;

# Loop over all the song data and update the records.
for songData in songDetails:

	# Let the user know which song we're working on.
	print "\nBuilding song data for \"" + songData["name"] + "\""

	# Each song needs to be written three times, once for the beginning of the song, once for the first third, and once for 2/3.
	for i in range(0, 3):

		# Let the user know which third we're working on.
		print "\n\t(" + str(i) + "/3)"

		# We need to calcualte the offset, depending on the position it the song. 0 will be 0/3 or just the beginning of the song, etc
		# The offset will be calculated by taking the global offset and adding a percentage of the file length to the global offset.
		offset = int( songData["globOffset"] + math.floor((i/3.0)*songData["length"]) )

		# The length needs to be adjusted based on the offset.
		# The first third (0/3) will just play the entire song, but the 1/3 needs to play only the last two thirds.
		length = int( songData["length"] - math.floor((i/3.0)*songData["length"]) )

		# Let the user know the offset and length for this third.
		print "\tOffset:" + str(offset) + "\n\tLength:" + str(length)

		# Write the offset for the current song to the header data.
		# (getIntBytes will always return 4 bytes with leading 00's!!)
		writeListAtPos(headerData, getIntBytes( offset ), filePos)

		# Write the length for the current song to the header data.
		# NOTE!: we offset filePos by 4 bytes.
		writeListAtPos(headerData, getIntBytes( length ), (filePos + 4) )

		# To write the song name, it needs to be a list (with whitespace built in at the end)
		# To build the list, we'll make a list of 16 spaces.
		songName = [" "]*16

		# Then write the name (which may be shorter than 16, but no longer than 16) to our song-name list.
		writeListAtPos(songName, list(songData["name16"]), 0)

		# ... and now we have a song name with approriate white space at the end!

		# Write the song name to the header data.
		# NOTE!: we offset filePos by 16 bytes because thats where the song name begins relative to this file's entry point.
		writeListAtPos(headerData, songName, (filePos + 16))

		# Now that we've written the critical data, we can move to the next 32 byte song block.
		filePos += 32

	#} NEXT i

#} NEXT songData


'''
	Okay, now we've added in the custom song data to our headers song index table.
	HOWEVER we should appened the footer block, and the song index table needs to be updated with those values as well.

	...

	What is the footer?

	The machine has a "programming mode," which acts like an automated phone call.
	It has six message mp3s and a "silence" mp3 that it uses to tell you how to configure the machine.

	I ripped these datafiles from the end of the MMC card that came with the machine.
	Since we aren't actually changing their data, their lengths (in bytes) will always remain constant.
	Their offsets will remain constant (relatively) as well.

	Here's a table that shows the offsets of the messages from the original card.
	The second column is their relative offsets:

	 File Name		offset		local offset (relative to the beginning of the messages)

		msg 1:		14472704	0
		msg 2:		14512128	39424
		msg 3:		14556160	83456
		msg 4:		14600192	127488
		msg 5:		14653440	180736
		msg 6:		14663680	190976
	  silence:		14707200	234496

  So... to build the table for these special messages we don't have to do much!

  All we need to change is their offsets, since their lengths and names will remain the same!

  And their offsets will just add the local column value to the end of the length of the songBlob and the header!
'''

# So let's do just that... write the offsets for the 6 messages and the silence.
# We can start by saving their relative offsets in an array.
msgRelativeOffsets = [0, 39424, 83456, 127488, 180736, 190976, 234496]

# Now we can loop over the array and add the offsets for the next 7 song blocks (which will always be the 6 messages and silence)
for offSet in msgRelativeOffsets:

	# Write the offset.
	# BTW, remember footerOffset? We calculated it at the end of appending the songs in ong big blob.
	writeListAtPos(headerData, getIntBytes( footerOffset + offSet ), filePos)

	# Increment filePos for the next iteration...
	filePos += 32

#} NEXT offSet

# Now that we've written the song data to the header, let's adjust the date stamp in the header!

# Get today's date in mm-dd-yyyy format and convert it to a list.
dateData = list( time.strftime("%m-%d-%Y") )

# Add it to the file.
# NOTE!: The date time stamp will always be at byte position 976 in the header data!
writeListAtPos(headerData, dateData, 976)

# Let's also change the playlist name from "Alaska 16" to "Custom 16"
# NOTE: the name will always be at position 998 in the header!
writeListAtPos(headerData, list("Custom 16 "), 998)

# Now that we've updated the header file, we can convert it back into a string.
headerData = "".join(headerData)

# We're almost done. Let's open up the footer data (which doesn't need to be modified at all) so we can build the final output.

# Open the footer file and read it's data.
footerFile = open('mmc_data/01-Footer', 'rb')
footerData = footerFile.read()
footerFile.close()

# Now we can assemble the output data. It's simply the headerData, the songBlob and the footerData appended together, in that order!
outputFileData = headerData + songBlob + footerData

# Let's open the output file for writing.
# Let the user know we're opening the file for write...
print "\nOpening file " + outputFileName + " for writing..."

# Open the the output file for writing.
outputFile = open(outputFileName, 'wb')

# Write the data to the output file.
outputFile.write(outputFileData)

# Close the output file and we're all done!
outputFile.close()

# Let the user know we're done!
# and give them some tips!
print "\n\nAll done!\n---------"
print "\nYou will need at least " + str( int( (len(outputFileData) / (1024.0*1024.0)) ) + 1 ) + " MB on the MMC card!"
print "\nUse command \"sudo dd if=" + outputFileName + " of=/dev/[path_to_your_MMC_card]\" to write the image to an MMC (not SD) card for the machine!"
print "\nSommerset Music Player Image Maker script by Greg Miller, 2014. www.gmiller.net"